# Copyright 2016-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for adapter ports.
"""

from __future__ import absolute_import

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS
from ._cmd_adapter import find_adapter


def find_port(cmd_ctx, client, cpc_or_name, adapter_name, port_name):
    """
    Find a port by name and return its resource object.
    """
    adapter = find_adapter(cmd_ctx, client, cpc_or_name, adapter_name)
    try:
        port = adapter.ports.find(name=port_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return port


@cli.group('port', options_metavar=COMMAND_OPTIONS_METAVAR)
def port_group():
    """
    Command group for managing adapter ports (DPM mode only).

    The commands in this group work only on CPCs that are in DPM mode.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@port_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('ADAPTER', type=str, metavar='ADAPTER')
@add_options(LIST_OPTIONS)
@click.pass_obj
def port_list(cmd_ctx, cpc, adapter, **options):
    """
    List the ports of an adapter.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_port_list(cmd_ctx, cpc, adapter, options))


@port_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('ADAPTER', type=str, metavar='ADAPTER')
@click.argument('PORT', type=str, metavar='PORT')
@click.pass_obj
def port_show(cmd_ctx, cpc, adapter, port):
    """
    Show the details of an adapter port.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Adapter.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_port_show(cmd_ctx, cpc, adapter, port))


@port_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('ADAPTER', type=str, metavar='ADAPTER')
@click.argument('PORT', type=str, metavar='PORT')
@click.option('--description', type=str, required=False,
              help='The new description of the port.')
@click.pass_obj
def port_update(cmd_ctx, cpc, adapter, port, **options):
    """
    Update the properties of an adapter port.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    The port may be on a physical adapter (e.g. a discovered OSA card) or a
    logical adapter (e.g. HiperSockets).

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_port_update(cmd_ctx, cpc, adapter, port,
                                                options))


def cmd_port_list(cmd_ctx, cpc_name, adapter_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    adapter = find_adapter(cmd_ctx, client, cpc_name, adapter_name)

    try:
        ports = adapter.ports.list(full_properties=True)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'name',
        'adapter',
        'cpc',
    ]
    if not options['names_only']:
        show_list.extend([
            'index',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    cpc_additions = {}
    adapter_additions = {}
    for port in ports:
        cpc_additions[port.uri] = cpc_name
        adapter_additions[port.uri] = adapter_name
    additions = {
        'adapter': adapter_additions,
        'cpc': cpc_additions,
    }

    try:
        print_resources(cmd_ctx, ports, cmd_ctx.output_format, show_list,
                        additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_port_show(cmd_ctx, cpc_name, adapter_name, port_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    port = find_port(cmd_ctx, client, cpc_name, adapter_name, port_name)

    try:
        port.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(port.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = adapter_name

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_port_update(cmd_ctx, cpc_name, adapter_name, port_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    port = find_port(cmd_ctx, client, cpc_name, adapter_name, port_name)

    org_options = original_options(options)
    properties = options_to_properties(org_options)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating port '{p}'.".
                   format(p=port_name))
        return

    try:
        port.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    # Adapter ports cannot be renamed.
    click.echo("Port '{p}' has been updated.".format(p=port_name))



# Copyright 2016-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for virtual functions.
"""

from __future__ import absolute_import

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS

from ._cmd_partition import find_partition


def find_vfunction(
        cmd_ctx, client, cpc_or_name, partition_name, vfunction_name):
    """
    Find a virtual function by name and return its resource object.
    """
    partition = find_partition(cmd_ctx, client, cpc_or_name, partition_name)
    try:
        vfunction = partition.virtual_functions.find(name=vfunction_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return vfunction


@cli.group('vfunction', options_metavar=COMMAND_OPTIONS_METAVAR)
def vfunction_group():
    """
    Command group for managing virtual functions (DPM mode only).

    The only virtual functions that can be managed with the commands in this
    group are those of the zEDC accelerator adapter, and only on CPCs that are
    in DPM mode.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@vfunction_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@add_options(LIST_OPTIONS)
@click.pass_obj
def vfunction_list(cmd_ctx, cpc, partition, **options):
    """
    List the virtual functions in a partition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_vfunction_list(cmd_ctx, cpc, partition,
                                                   options))


@vfunction_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('VFUNCTION', type=str, metavar='VFUNCTION')
@click.pass_obj
def vfunction_show(cmd_ctx, cpc, partition, vfunction):
    """
    Show the details of a virtual function.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Partition.
      - 'adapter-name' - Name of the Adapter referenced by 'adapter-uri'.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_vfunction_show(cmd_ctx, cpc, partition,
                                                   vfunction))


@vfunction_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.option('--name', type=str, required=True,
              help='The name of the new virtual function. Must be unique '
              'within the virtual functions of the partition')
@click.option('--description', type=str, required=False,
              help='The description of the new virtual function. '
              'Default: empty')
@click.option('--adapter', type=str, required=True,
              help='The name of the adapter backing the virtual function.')
@click.option('--device-number', type=str, required=False,
              help='The device number to be used for the new virtual '
              'function. Default: auto-generated')
@click.pass_obj
def vfunction_create(cmd_ctx, cpc, partition, **options):
    """
    Create a virtual function in a partition.

    This assigns a virtual function of an adapter to a partition, creating
    a named virtual function resource within that partition.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_vfunction_create(cmd_ctx, cpc, partition,
                                                     options))


@vfunction_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('VFUNCTION', type=str, metavar='VFUNCTION')
@click.option('--name', type=str, required=False,
              help='The new name of the virtual function. Must be unique '
              'within the virtual functions of the partition.')
@click.option('--description', type=str, required=False,
              help='The new description of the virtual function.')
@click.option('--adapter', type=str, required=False,
              help='The name of the new adapter (in the same CPC) that will '
              'back the virtual function.')
@click.option('--device-number', type=str, required=False,
              help='The new device number to be used for the virtual '
              'function.')
@click.pass_obj
def vfunction_update(cmd_ctx, cpc, partition, vfunction, **options):
    """
    Update the properties of a virtual function.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_vfunction_update(cmd_ctx, cpc, partition,
                                                     vfunction, options))


@vfunction_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION', type=str, metavar='PARTITION')
@click.argument('VFUNCTION', type=str, metavar='VFUNCTION')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the virtual function.',
              prompt='Are you sure you want to delete this virtual function ?')
@click.pass_obj
def vfunction_delete(cmd_ctx, cpc, partition, vfunction):
    """
    Delete a virtual function.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_vfunction_delete(cmd_ctx, cpc, partition,
                                                     vfunction))


def cmd_vfunction_list(cmd_ctx, cpc_name, partition_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)

    try:
        vfunctions = partition.virtual_functions.list()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'name',
        'cpc',
        'partition',
    ]
    if not options['names_only']:
        show_list.extend([
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    cpc_additions = {}
    partition_additions = {}
    for vfunction in vfunctions:
        cpc_additions[vfunction.uri] = cpc_name
        partition_additions[vfunction.uri] = partition_name
    additions = {
        'cpc': cpc_additions,
        'partition': partition_additions,
    }

    try:
        print_resources(cmd_ctx, vfunctions, cmd_ctx.output_format, show_list,
                        additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_vfunction_show(cmd_ctx, cpc_name, partition_name, vfunction_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    vfunction = find_vfunction(cmd_ctx, client, cpc_name, partition_name,
                               vfunction_name)

    try:
        vfunction.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(vfunction.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = partition_name

    # Add artificial property 'adapter-name'
    adapter_uri = vfunction.get_property('adapter-uri')
    adapter_props = client.session.get(adapter_uri)
    adapter_name = adapter_props['name']
    properties['adapter-name'] = adapter_name

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_vfunction_create(cmd_ctx, cpc_name, partition_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    partition = find_partition(cmd_ctx, client, cpc_name, partition_name)

    name_map = {
        # The following options are handled in this function:
        'adapter': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    adapter_name = org_options['adapter']
    try:
        adapter = partition.manager.cpc.adapters.find(name=adapter_name)
    except zhmcclient.NotFound:
        raise click_exception("Could not find adapter '{a}' in CPC '{c}'.".
                              format(a=adapter_name, c=cpc_name),
                              cmd_ctx.error_format)
    properties['adapter-uri'] = adapter.uri

    try:
        new_vfunction = partition.virtual_functions.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New virtual function '{f}' has been created.".
               format(f=new_vfunction.properties['name']))


def cmd_vfunction_update(cmd_ctx, cpc_name, partition_name, vfunction_name,
                         options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    vfunction = find_vfunction(cmd_ctx, client, cpc_name, partition_name,
                               vfunction_name)

    name_map = {
        # The following options are handled in this function:
        'adapter': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    if org_options['adapter'] is not None:
        adapter_name = org_options['adapter']
        try:
            adapter = vfunction.manager.partition.manager.cpc.adapters.find(
                name=adapter_name)
        except zhmcclient.NotFound:
            raise click_exception("Could not find adapter '{a}' in CPC '{c}'.".
                                  format(a=adapter_name, c=cpc_name),
                                  cmd_ctx.error_format)
        properties['adapter-uri'] = adapter.uri

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating virtual function "
                   "{f}.".format(f=vfunction_name))
        return

    try:
        vfunction.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    if 'name' in properties and properties['name'] != vfunction_name:
        click.echo("Virtual function '{f}' has been renamed to '{fn}' and was "
                   "updated.".format(f=vfunction_name, fn=properties['name']))
    else:
        click.echo("Virtual function '{f}' has been updated.".
                   format(f=vfunction_name))


def cmd_vfunction_delete(cmd_ctx, cpc_name, partition_name, vfunction_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    vfunction = find_vfunction(cmd_ctx, client, cpc_name, partition_name,
                               vfunction_name)

    try:
        vfunction.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Virtual function '{f}' has been deleted.".
               format(f=vfunction_name))



# Copyright 2016-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for virtual switches.
"""

from __future__ import absolute_import

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS
from ._cmd_cpc import find_cpc


def find_vswitch(cmd_ctx, client, cpc_or_name, vswitch_name):
    """
    Find a virtual switch by name and return its resource object.
    """
    if isinstance(cpc_or_name, zhmcclient.Cpc):
        cpc = cpc_or_name
    else:
        cpc = find_cpc(cmd_ctx, client, cpc_or_name)
    # The CPC must be in DPM mode. We don't check that because it would
    # cause a GET to the CPC resource that we otherwise don't need.
    try:
        vswitch = cpc.virtual_switches.find(name=vswitch_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return vswitch


@cli.group('vswitch', options_metavar=COMMAND_OPTIONS_METAVAR)
def vswitch_group():
    """
    Command group for managing virtual switches (DPM mode only).

    Virtual switches are automatically established by the system for OSA
    and Hipersocket adapters. They do not exist for ROCE and CNA adapters.

    The commands in this group work only on CPCs that are in DPM mode.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@vswitch_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@add_options(LIST_OPTIONS)
@click.option('--adapter', is_flag=True, required=False, hidden=True)
@click.pass_obj
def vswitch_list(cmd_ctx, cpc, **options):
    """
    List the virtual switches in a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_vswitch_list(cmd_ctx, cpc, options))


@vswitch_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('VSWITCH', type=str, metavar='VSWITCH')
@click.pass_obj
def vswitch_show(cmd_ctx, cpc, vswitch):
    """
    Show the details of a virtual switch.

    The following properties are shown in addition to those returned by the HMC:

    \b
    - 'parent-name' - Name of the parent CPC.
    - 'backing-adapter-name' - Name of the Network Adapter referenced by
      'backing-adapter-uri'.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_vswitch_show(cmd_ctx, cpc, vswitch))


@vswitch_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('VSWITCH', type=str, metavar='VSWITCH')
@click.option('--name', type=str, required=False,
              help='The new name of the virtual switch.')
@click.option('--description', type=str, required=False,
              help='The new description of the virtual switch.')
@click.pass_obj
def vswitch_update(cmd_ctx, cpc, vswitch, **options):
    """
    Update the properties of a virtual switch.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_vswitch_update(cmd_ctx, cpc, vswitch,
                                                   options))


def cmd_vswitch_list(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    try:
        vswitches = cpc.virtual_switches.list()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if options['adapter']:
        click.echo("The --adapter option is deprecated and adapter information "
                   "is now always shown.")

    show_list = [
        'name',
        'cpc',
    ]
    if not options['names_only']:
        show_list.extend([
            'adapter',
            'type',
            'port',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])

    cpc_additions = {}
    adapter_additions = {}
    for vswitch in vswitches:
        cpc_additions[vswitch.uri] = cpc_name
        try:
            adapter_uri = vswitch.prop('backing-adapter-uri')
            adapter = cpc.adapters.find(**{'object-uri': adapter_uri})
            adapter_additions[vswitch.uri] = adapter.name
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
    additions = {
        'cpc': cpc_additions,
        'adapter': adapter_additions,
    }

    try:
        print_resources(cmd_ctx, vswitches, cmd_ctx.output_format, show_list,
                        additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_vswitch_show(cmd_ctx, cpc_name, vswitch_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    vswitch = find_vswitch(cmd_ctx, client, cpc_name, vswitch_name)

    try:
        vswitch.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(vswitch.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = cpc_name

    # Add artificial property 'backing-adapter-name'
    adapter_uri = vswitch.get_property('backing-adapter-uri')
    adapter_props = client.session.get(adapter_uri)
    properties['backing-adapter-name'] = adapter_props['name']

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_vswitch_update(cmd_ctx, cpc_name, vswitch_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    vswitch = find_vswitch(cmd_ctx, client, cpc_name, vswitch_name)

    org_options = original_options(options)
    properties = options_to_properties(org_options)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating virtual switch '{s}'.".
                   format(s=vswitch_name))
        return

    try:
        vswitch.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    if 'name' in properties and properties['name'] != vswitch_name:
        click.echo("Virtual switch '{s}' has been renamed to '{sn}' and was "
                   "updated.".format(s=vswitch_name, sn=properties['name']))
    else:
        click.echo("Virtual switch '{s}' has been updated.".
                   format(s=vswitch_name))



# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for the HMC console.
"""

from __future__ import absolute_import

import click
import dateutil.parser

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_dicts, print_list, \
    TABLE_FORMATS, hide_property, COMMAND_OPTIONS_METAVAR, click_exception


@cli.group('console', options_metavar=COMMAND_OPTIONS_METAVAR)
def console_group():
    """
    Command group for the HMC console.

    These commands always target the (one) console of the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@console_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--all', is_flag=True, required=False,
              help='Show all properties. Default: Hide some properties in '
              'table output formats')
@click.pass_obj
def console_show(cmd_ctx, **options):
    """
    Show properties of the console of the targeted HMC.

    \b
    In table output formats, the following properties are hidden by default
    but can be shown by using the --all option:
      - ec-mcl-description
      - network-info

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_show(cmd_ctx, options))


@console_group.command('restart', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--force', '-f', is_flag=True, required=False,
              help='Perform the restart regardless of whether users are '
              'logged in to the HMC. Users in this sense are local or remote '
              'GUI users. Disconnected GUI users and users at the WS-API do '
              'not count for this.')
@click.option('--wait', '-w', is_flag=True, required=False,
              help='Wait for the HMC to become available again. Note that '
              'any session IDs have become invalid.')
@click.option('--timeout', '-T', type=int, required=False, default=60,
              help='Timeout (in seconds) when waiting for the HMC to become '
              'available again. Default: 60.')
@click.pass_obj
def console_restart(cmd_ctx, **options):
    """
    Restart the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_restart(cmd_ctx, options))


@console_group.command('get-audit-log', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--begin', type=str, required=False,
              help='Begin time for the log, in a format supported by '
              'the python-dateutil package. Default: Earliest available.')
@click.option('--end', type=str, required=False,
              help='End time for the log, in a format supported by '
              'the python-dateutil package. Default: Latest available.')
@click.pass_obj
def get_audit_log(cmd_ctx, **options):
    """
    Get the audit log of the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_get_audit_log(cmd_ctx, options))


@console_group.command('get-security-log',
                       options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--begin', type=str, required=False,
              help='Begin time for the log, in a format supported by '
              'the python-dateutil package. Default: Earliest available.')
@click.option('--end', type=str, required=False,
              help='End time for the log, in a format supported by '
              'the python-dateutil package. Default: Latest available.')
@click.pass_obj
def get_security_log(cmd_ctx, **options):
    """
    Get the security log of the targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_get_security_log(cmd_ctx, options))


@console_group.command('list-api-features',
                       options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--name', type=str, metavar='NAME',
              required=False,
              help='A regular expression used to limit returned objects to '
                   'those that have a matching name field.')
@click.pass_obj
def list_api_features(cmd_ctx, **options):
    """
    Lists the Web Services API features available of the console of the
    targeted HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_list_api_features(cmd_ctx, options))


@console_group.command('upgrade', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--bundle-level', '-b', type=str, required=True,
              help="Name of the bundle to be installed on the HMC "
              "(e.g. 'H71').")
@click.option('--backup-location-type', type=str, required=False, default='usb',
              help="Type of backup location for the HMC backup that is "
              "performed: 'ftp': The FTP server that was used for the last "
              "console backup as defined on the 'Configure Backup Settings' "
              "user interface task in the HMC GUI., "
              "'usb': The USB storage device mounted to the HMC.")
@click.option('--accept-firmware', '-a', type=bool, required=False,
              default=True,
              help="Boolean indicating to accept the previous bundle level "
              "before installing the new level. Default: true")
@click.option('--timeout', '-T', type=int, required=False, default=1200,
              help='Timeout (in seconds) when waiting for the HMC upgrade '
              'to be complete. Default: 1200.')
@click.pass_obj
def console_upgrade(cmd_ctx, **options):
    """
    Upgrade the firmware on the targeted HMC to a new bundle level.

    This is done by performing the "Console Single Step Install" operation
    which performs the following steps:

    \b
    * If `accept-firmware` is True, the firmware currently installed on the
      this HMC is accepted. Note that once firmware is accepted, it cannot be
      removed.
    * A backup of the this HMC is performed to the specified backup device.
    * The new firmware identified by the bundle-level field is retrieved from
      the IBM support site and installed.
    * The newly installed firmware is activated, which includes rebooting this
      HMC.

    Note that it is not possible to downgrade the HMC firmware with this
    operation.

    If the HMC firmware is already at the requested bundle level, nothing is
    changed and the command succeeds.

    For HMCs that run on an HMA that also hosts an SE (e.g. z16 and higher),
    the HMC firmware can only be upgraded if the HMA hosts an alternate SE.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_console_upgrade(cmd_ctx, options))


def cmd_console_show(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    try:
        console.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(console.properties)

    # Hide some long or deeply nested properties in table output formats.
    if not options['all'] and cmd_ctx.output_format in TABLE_FORMATS:
        hide_property(properties, 'ec-mcl-description')
        hide_property(properties, 'network-info')

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_console_restart(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    force = options['force']
    wait = options['wait']
    timeout = options['timeout']

    if wait:
        click.echo("Restarting the HMC and waiting for its availability "
                   "(timeout: {} s)".format(timeout))
    else:
        click.echo("Restarting the HMC")
    try:
        console.restart(force=force, wait_for_available=wait,
                        operation_timeout=timeout)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    if wait:
        click.echo("The HMC has restarted and is available again. Any earlier "
                   "session IDs have become invalid")
    else:
        click.echo("The HMC is currently restarting. Wait for it to be "
                   "available again before issuing any commands. Any earlier "
                   "session IDs will be invalid")


def cmd_get_audit_log(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    begin_time = options['begin'] or None
    if begin_time:
        begin_time = dateutil.parser.parse(begin_time)
    end_time = options['end'] or None
    if end_time:
        end_time = dateutil.parser.parse(end_time)

    try:
        log_items = console.get_audit_log(begin_time, end_time)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'timestamp',
        'event-name',
        'event-id',
        'userid',
        'event-message',
    ]

    ts_additions = {}
    for i, log_item in enumerate(log_items):
        hmc_ts = log_item['event-time']
        dt = zhmcclient.datetime_from_timestamp(hmc_ts)
        ts_additions[i] = str(dt)
    additions = {
        'timestamp': ts_additions,
    }

    cmd_ctx.spinner.stop()
    print_dicts(cmd_ctx, log_items, cmd_ctx.output_format,
                show_list=show_list, additions=additions, all=False)


def cmd_get_security_log(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    begin_time = options['begin'] or None
    if begin_time:
        begin_time = dateutil.parser.parse(begin_time)
    end_time = options['end'] or None
    if end_time:
        end_time = dateutil.parser.parse(end_time)

    try:
        log_items = console.get_security_log(begin_time, end_time)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'timestamp',
        'event-name',
        'event-id',
        'userid',
        'event-message',
    ]

    ts_additions = {}
    for i, log_item in enumerate(log_items):
        hmc_ts = log_item['event-time']
        dt = zhmcclient.datetime_from_timestamp(hmc_ts)
        ts_additions[i] = str(dt)
    additions = {
        'timestamp': ts_additions,
    }

    cmd_ctx.spinner.stop()
    print_dicts(cmd_ctx, log_items, cmd_ctx.output_format,
                show_list=show_list, additions=additions, all=False)


def cmd_list_api_features(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    name = options['name']
    try:
        features = console.list_api_features(name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    print_list(cmd_ctx, features, cmd_ctx.output_format)


def cmd_console_upgrade(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    bundle_level = options['bundle_level']
    accept_firmware = options['accept_firmware']
    backup_location_type = options['backup_location_type']
    timeout = options['timeout']

    ec_mcl = console.prop('ec-mcl-description')
    hmc_bundle_level = ec_mcl.get('bundle-level', None)
    if hmc_bundle_level is None:
        hmc_version = console.prop('version')
        raise click_exception(
            "HMC version {v} does not support firmware upgrade through "
            "the Web Services API".format(v=hmc_version),
            cmd_ctx.error_format)

    click.echo("Upgrading the HMC to bundle level {bl} and waiting for "
               "completion (timeout: {t} s)".
               format(bl=bundle_level, t=timeout))

    try:
        console.single_step_install(
            bundle_level=bundle_level,
            accept_firmware=accept_firmware,
            backup_location_type=backup_location_type,
            wait_for_completion=True,
            operation_timeout=timeout)
    except zhmcclient.HTTPError as exc:
        if exc.http_status == 400 and exc.reason == 356:
            # HMC was already at that bundle level
            cmd_ctx.spinner.stop()
            click.echo("The HMC was already at bundle level {bl} and did "
                       "not need to be changed".
                       format(bl=bundle_level))
            return
        raise click_exception(exc, cmd_ctx.error_format)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("The HMC has been upgraded to bundle level {bl} and is "
               "available again. Any earlier session IDs have become invalid".
               format(bl=bundle_level))



# Copyright 2016-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for CPCs.
"""

from __future__ import absolute_import

import io
import json
import yaml
import yamlloader
import click
from tabulate import tabulate

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, print_list, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, TABLE_FORMATS, hide_property, \
    required_option, abort_if_false, validate, print_dicts


POWER_SAVING_TYPES = ['high-performance', 'low-power', 'custom']
DEFAULT_POWER_SAVING_TYPE = 'high-performance'
POWER_CAPPING_STATES = ['disabled', 'enabled', 'custom']

# Data formats of DPM configuration file
DPM_FORMATS = ['yaml', 'json']
DEFAULT_DPM_FORMAT = 'yaml'


# JSON schema for adapter mapping file
MAPPING_SCHEMA = {
    "title": "adapter mapping file schema",
    "description":
        "JSON schema that defines the structure of an adapter mapping file.",
    "type": "object",
    "additionalProperties": False,
    "required": [
        "adapter-mapping"
    ],
    "properties": {
        "adapter-mapping": {
            "description":
                "List of PCHID mappings from config file to import into CPC.",
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": False,
                "required": [
                    "old-adapter-id",
                    "new-adapter-id",
                ],
                "properties": {
                    "old-adapter-id": {
                        "description":
                            "PCHID of adapter in DPM configuration file.",
                        "type": "string"
                    },
                    "new-adapter-id": {
                        "description":
                            "PCHID of adapter when imported into the CPC.",
                        "type": "string"
                    }
                }
            }
        }
    }
}


def help_dpm_file(cmd_ctx, param, value):
    # pylint: disable=unused-argument
    """
    Click callback function for --help-dpm-file option, that displays
    help for the format of the DPM configuration file and exits.
    """
    if not value or cmd_ctx.resilient_parsing:
        return
    print("""
Format of DPM configuration file:

The DPM configuration file contains the properties of a single CPC and all of
its child objects (such as partitions or adapters) and all of its CPC-specific
related objects (such as storage groups associated with the CPC).

The DPM configuration file is written by the 'zhmc cpc dpm-export' command
and read by the 'zhmc cpc dpm-import' command.

The DPM configuration file is in YAML or JSON format and has the structure of
the payload of the 'Import DPM Configuration' operation described in the
HMC API book, except for the following properties which are not in that file:

* 'preserve-uris' - Boolean controlling whether to preserve object URIs and IDs.

* 'preserve-wwpns' - Boolean controlling whether to preserve HBA WWPNs.

* 'adapter-mapping' - List of mappings of adapter PCHIDs between the CPC from
  which the DPM configuration was exported and the new CPC to which it is
  imported.

Instead of being taken from the DPM configuration file, the
'zhmc cpc dpm-import' command sets these properties based on the
corresponding options when issuing the 'Import DPM Configuration' operation.
""")
    cmd_ctx.exit()


def help_mapping_file(cmd_ctx, param, value):
    # pylint: disable=unused-argument
    """
    Click callback function for --help-dpm-file option, that displays
    help for the format of the adapter mapping file and exits.
    """
    if not value or cmd_ctx.resilient_parsing:
        return
    print("""
Format of adapter mapping file:

The adapter mapping file specifies how PCHIDs of adapters in a DPM
configuration file need to be replaced when importing the DPM configuration
into a new CPC. If you do not provide a mapping for an adapter, DPM uses a
one-to-one mapping of adapters in the configuration file to adapters on the
target system.

The adapter mapping file is created manually and is read by the
'zhmc cpc dpm-import' command.

The adapter mapping file is in YAML format and has the following structure:

    adapter-mapping:
      - old-adapter-id:  A11  # PCHID in the DPM configuration file
        new-adapter-id: "911" # PCHID in the new system that maps to A11
      - old-adapter-id: "11c"
        new-adapter-id: "12C"
      - ... # More mappings, one for each adapter PCHID you intend to map

The values of the 'old-adapter-id' and 'new-adapter-id' properties are the
PCHID values as a hexadecimal string. Therefore, hexadecimal values that consist
only of decimal digits (like '911' in the example above) must be put into double
quotes. The characters in the hexadecimal string may be in upper or lower case.
""")
    cmd_ctx.exit()


def find_cpc(cmd_ctx, client, cpc_name):
    """
    Find a CPC by name and return its resource object.
    """
    try:
        cpc = client.cpcs.find(name=cpc_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return cpc


@cli.group('cpc', options_metavar=COMMAND_OPTIONS_METAVAR)
def cpc_group():
    """
    Command group for managing CPCs.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@cpc_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--type', is_flag=True, required=False, hidden=True)
@click.option('--mach', is_flag=True, required=False, hidden=True)
@add_options(LIST_OPTIONS)
@click.pass_obj
def cpc_list(cmd_ctx, **options):
    """
    List the CPCs managed by the HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_list(cmd_ctx, options))


@cpc_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--all', is_flag=True, required=False,
              help='Show all properties. Default: Hide some properties in '
              'table output formats')
@click.pass_obj
def cpc_show(cmd_ctx, cpc, **options):
    """
    Show details of a CPC.

    \b
    In table output formats, the following properties are hidden by default
    but can be shown by using the --all option:
      - auto-start-list
      - available-features-list
      - cpc-power-saving-state
      - ec-mcl-description
      - network1-ipv6-info
      - network2-ipv6-info
      - stp-configuration

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_show(cmd_ctx, cpc, options))


@cpc_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--description', type=str, required=False,
              help='The new description of the CPC. '
              '(DPM mode only).')
@click.option('--acceptable-status', type=str, required=False,
              help='The new set of acceptable operational status values, as a '
              'comma-separated list. The empty string specifies an empty list.')
@click.option('--next-activation-profile', type=str, required=False,
              help='The name of the new next reset activation profile. '
              '(not in DPM mode).')
@click.option('--processor-time-slice', type=int, required=False,
              help='The new time slice (in ms) for logical processors. '
              'A value of 0 causes the time slice to be dynamically '
              'determined by the system. A positive value causes a constant '
              'time slice to be used. '
              '(not in DPM mode).')
@click.option('--wait-ends-slice/--no-wait-ends-slice', default=None,
              required=False,
              help='The new setting for making logical processors lose their '
              'time slice when they enter a wait state. '
              '(not in DPM mode).')
@click.pass_obj
def cpc_update(cmd_ctx, cpc, **options):
    """
    Update the properties of a CPC.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_update(cmd_ctx, cpc, options))


@cpc_group.command('set-power-save', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--power-saving', type=click.Choice(POWER_SAVING_TYPES),
              required=False, default=DEFAULT_POWER_SAVING_TYPE,
              help='Defines the type of power saving. Default: {pd}'.
              format(pd=DEFAULT_POWER_SAVING_TYPE))
@click.pass_obj
def set_power_save(cmd_ctx, cpc, **options):
    """
    Set the power save settings of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_set_power_save(cmd_ctx, cpc, options))


@cpc_group.command('set-power-capping',
                   options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--power-capping-state', type=click.Choice(POWER_CAPPING_STATES),
              required=True,
              help='Defines the state of power capping.')
@click.option('--power-cap-current', type=int, required=False,
              help='Specifies the current cap value for the CPC in watts (W). '
              'Required if power capping state is enabled.')
@click.pass_obj
def set_power_capping(cmd_ctx, cpc, **options):
    """
    Set the power capping settings of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_set_power_capping(cmd_ctx, cpc,
                                                          options))


@cpc_group.command('get-em-data', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.pass_obj
def get_em_data(cmd_ctx, cpc):
    """
    Get all energy management data of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_get_em_data(cmd_ctx, cpc))


@cpc_group.command('dpm-export', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC', required=True)
@click.option('--dpm-file', '-d', metavar='FILE', type=str, required=True,
              help='Path name of the DPM configuration file to be written.')
@click.option('--dpm-format', '-f', type=click.Choice(DPM_FORMATS),
              required=False, default=DEFAULT_DPM_FORMAT,
              help='Controls in which data format the DPM configuration file '
              'is written. Default: {d}.'.format(d=DEFAULT_DPM_FORMAT))
@click.option('--help-dpm-file', is_flag=True, is_eager=True,
              callback=help_dpm_file, expose_value=False,
              help='Show help on the format of the DPM configuration file '
              'and exit.')
@click.option('--include-unused-adapters', '-i', is_flag=True, default=False,
              help='Controls whether the full set of adapters should be '
              'returned, vs. only those that are referenced by other DPM '
              'objects that are part of the return data. Default: not included')
@click.pass_obj
def dpm_export(cmd_ctx, cpc, **options):
    """
    Export a DPM configuration from a CPC.

    The DPM configuration of the CPC is exported and written to a DPM
    configuration file (in YAML or JSON format).
    For details about the format of that file, use --help-dpm-file.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_dpm_export(cmd_ctx, cpc, options))


@cpc_group.command('dpm-import', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC', required=True)
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm import of the DPM configuration.',
              prompt='Are you sure you want to import the DPM configuration '
              'into the CPC, replacing its current DPM configuration ?')
@click.option('--dpm-file', '-d', metavar='FILE', type=str, required=True,
              help='Path name of the DPM configuration file to be used.')
@click.option('--dpm-format', '-f', type=click.Choice(DPM_FORMATS),
              required=False, default=DEFAULT_DPM_FORMAT,
              help='Controls in which data format the DPM configuration file '
              'is read. Default: {d}.'.format(d=DEFAULT_DPM_FORMAT))
@click.option('--preserve-uris', is_flag=True, required=False,
              help='Controls whether the URIs and IDs of objects in the DPM '
              'configuration file are preserved or ignored and newly '
              'generated. Default: Ignored and newly generated.')
@click.option('--preserve-wwpns', is_flag=True, required=False,
              help='Controls whether the WWPNs of HBAs in the DPM '
              'configuration file are preserved or ignored and newly '
              'generated. Default: Ignored and newly generated.')
@click.option('--mapping-file', '-m', metavar='FILE', type=str, required=False,
              help='Path name of the adapter mapping file to be used. '
              'Default: Use 1:1 adapter mapping.')
@click.option('--help-dpm-file', is_flag=True, is_eager=True,
              callback=help_dpm_file, expose_value=False,
              help='Show help on the format of the DPM configuration file '
              'and exit.')
@click.option('--help-mapping-file', is_flag=True, is_eager=True,
              callback=help_mapping_file, expose_value=False,
              help='Show help on the format of the adapter mapping file '
              'and exit.')
@click.pass_obj
def dpm_import(cmd_ctx, cpc, **options):
    """
    Import a DPM configuration into a CPC.

    The DPM configuration of the CPC is read from a DPM configuration file
    (in YAML or JSON format) and imported into the CPC, replacing the current
    DPM configuration of the CPC.
    For details about the format of that file, use --help-dpm-file.

    Optionally, an adapter mapping file (in YAML format) can be specified that
    can be used to accomodate different adapter PCHIDs (plug positions) between
    the CPC represented in the DPM configuration and the CPC targeted for the
    import. By default, the adapters are assumed to be in the same plug
    positions, i.e. the PCHIDs will be unchanged.
    For details about the format of that file, use --help-mapping-file.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_dpm_import(cmd_ctx, cpc, options))


@cpc_group.group('autostart', options_metavar=COMMAND_OPTIONS_METAVAR)
def cpc_autostart_group():
    """
    Command group for managing the auto-start list of a CPC (in DPM mode).

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@cpc_autostart_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.pass_obj
def cpc_autostart_show(cmd_ctx, cpc):
    """
    Show the auto-start list of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_autostart_show(cmd_ctx, cpc))


@cpc_autostart_group.command('add', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITIONS_DELAY', type=(str, int), metavar='PARTITIONS DELAY')
@click.option('--group', type=str, metavar='GROUP',
              required=False,
              help='Add the partition(s) as a partition group with this name. '
              'Required when adding a group.')
@click.option('--description', type=str, metavar='TEXT',
              required=False,
              help='Description of partition group. '
              'Default: No description.')
@click.option('--before', type=str, metavar='PARTITION_OR_GROUP',
              required=False,
              help='Insert the new partition or group before this '
              'partition/group. '
              'Default: Append new partition or group to the end.')
@click.option('--after', type=str, metavar='PARTITION_OR_GROUP',
              required=False,
              help='Insert the new partition or group after this '
              'partition/group. '
              'Default: Append new partition or group to the end.')
@click.pass_obj
def cpc_autostart_add(cmd_ctx, cpc, partitions_delay, **options):
    """
    Add a partition or group to the auto-start list of a CPC.

    A partition group exists only in context of the auto-start list; it has
    nothing to do with Group objects.

    PARTITIONS is the partition name or a comma-separated list of partition
    names in case of adding a partition group.

    DELAY is the delay afer starting this partition or group, in seconds.

    The updated auto-start list is shown.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_autostart_add(cmd_ctx, cpc, partitions_delay, options))


@cpc_autostart_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('PARTITION_OR_GROUP', type=str, metavar='PARTITION_OR_GROUP')
@click.pass_obj
def cpc_autostart_delete(cmd_ctx, cpc, partition_or_group):
    """
    Delete a partition or group from the auto-start list of a CPC.

    The updated auto-start list is shown.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_autostart_delete(cmd_ctx, cpc, partition_or_group))


@cpc_autostart_group.command('clear', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.pass_obj
def cpc_autostart_clear(cmd_ctx, cpc):
    """
    Clear the auto-start list of a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_autostart_clear(cmd_ctx, cpc))


@cpc_group.command('list-api-features', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--name', type=str, metavar='NAME',
              required=False,
              help='A regular expression used to limit returned objects to '
                   'those that have a matching name field.')
@click.pass_obj
def cpc_list_api_features(cmd_ctx, cpc, **options):
    """
    Lists the Web Services API features available on a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_cpc_list_api_features(cmd_ctx, cpc, options))


@cpc_group.command('upgrade', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--bundle-level', '-b', type=str, required=True,
              help="Name of the bundle to be installed on the SE "
              "(e.g. 'S71').")
@click.option('--accept-firmware', '-a', type=bool, required=False,
              default=True,
              help="Boolean indicating to accept the previous bundle level "
              "before installing the new level. Default: true")
@click.option('--timeout', '-T', type=int, required=False, default=1200,
              help='Timeout (in seconds) when waiting for the SE upgrade '
              'to be complete. Default: 1200.')
@click.pass_obj
def cpc_upgrade(cmd_ctx, cpc, **options):
    """
    Upgrade the firmware on the Support Element (SE) of a CPC to a new bundle
    level.

    This is done by performing the "CPC Single Step Install" operation
    which performs the following steps:

    \b
    * A backup of the target CPC is performed to its SE hard drive.
    * If `accept_firmware` is True, the firmware currently installed on the SE
      of this CPC is accepted. Note that once firmware is accepted, it cannot be
      removed.
    * The new firmware identified by the bundle-level field is retrieved from
      the IBM support site and installed.
    * The newly installed firmware is activated, which includes rebooting the SE
      of this CPC.

    If an error occurrs when installing the upgrades for the components of
    the new bundle, any components that were successfully installed are
    rolled back.

    If an error occurs after the firmware is accepted, the firmware remains
    accepted.

    Note that it is not possible to downgrade the SE firmware with this
    operation.

    If the SE firmware is already at the requested bundle level, nothing is
    changed and the command succeeds.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_cpc_upgrade(cmd_ctx, cpc, options))


def cmd_cpc_list(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)

    try:
        cpcs = client.cpcs.list()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if options['type']:
        click.echo("The --type option is deprecated and type information "
                   "is now always shown.")
    if options['mach']:
        click.echo("The --mach option is deprecated and machine information "
                   "is now always shown.")

    show_list = [
        'name',
    ]
    if not options['names_only']:
        show_list.extend([
            'status',
            'dpm-enabled',
            'se-version',
            'machine-type',
            'machine-model',
            'machine-serial-number',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])

    try:
        print_resources(cmd_ctx, cpcs, cmd_ctx.output_format, show_list,
                        all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_cpc_show(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    try:
        cpc.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(cpc.properties)

    # Hide some long or deeply nested properties in table output formats.
    if not options['all'] and cmd_ctx.output_format in TABLE_FORMATS:
        hide_property(properties, 'auto-start-list')
        hide_property(properties, 'available-features-list')
        hide_property(properties, 'cpc-power-saving-state')
        hide_property(properties, 'ec-mcl-description')
        hide_property(properties, 'network1-ipv6-info')
        hide_property(properties, 'network2-ipv6-info')
        hide_property(properties, 'stp-configuration')

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_cpc_update(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    name_map = {
        'next-activation-profile': 'next-activation-profile-name',
        'processor-time-slice': None,
        'wait-ends-slice': None,
        'no-wait-ends-slice': None,
        'acceptable-status': None,
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    time_slice = org_options['processor-time-slice']
    if time_slice is None:
        # 'processor-running-time*' properties not changed
        pass
    elif time_slice < 0:
        raise click_exception("Value for processor-time-slice option must "
                              "be >= 0", cmd_ctx.error_format)
    elif time_slice == 0:
        properties['processor-running-time-type'] = 'system-determined'
    else:  # time_slice > 0
        properties['processor-running-time-type'] = 'user-determined'
        properties['processor-running-time'] = time_slice

    if org_options['wait-ends-slice'] is not None:
        properties['does-wait-state-end-time-slice'] = \
            org_options['wait-ends-slice']

    if org_options['acceptable-status'] is not None:
        status_list = org_options['acceptable-status'].split(',')
        status_list = [item for item in status_list if item]
        properties['acceptable-status'] = status_list

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating CPC '{c}'.".
                   format(c=cpc_name))
        return

    try:
        cpc.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()

    # Name changes are not supported for CPCs.
    click.echo("CPC '{c}' has been updated.".format(c=cpc_name))


def cmd_cpc_set_power_save(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    org_options = original_options(options)
    power_saving = org_options['power-saving']
    cpc.set_power_save(power_saving)
    cmd_ctx.spinner.stop()
    click.echo("The power save settings of CPC '{c}' have been set to {s}.".
               format(c=cpc_name, s=power_saving))


def cmd_cpc_set_power_capping(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    org_options = original_options(options)
    power_capping_state = org_options['power-capping-state']
    power_cap_current = None
    if org_options['power-cap-current']:
        power_cap_current = org_options['power-cap-current']
    cpc.set_power_capping(org_options['power-capping-state'], power_cap_current)
    cmd_ctx.spinner.stop()
    click.echo("The power capping settings of CPC '{c}' have been set to {s}.".
               format(c=cpc_name, s=power_capping_state))


def cmd_cpc_get_em_data(cmd_ctx, cpc_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    energy_props = cpc.get_energy_management_properties()
    print_properties(cmd_ctx, energy_props, cmd_ctx.output_format)


def cmd_dpm_export(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    dpm_file = required_option(options, 'dpm_file')
    dpm_format = options['dpm_format']

    include_unused_adapters = options['include_unused_adapters']

    config_dict = cpc.export_dpm_configuration(include_unused_adapters)
    try:
        with io.open(dpm_file, 'w', encoding='utf-8') as fp:
            if dpm_format == 'yaml':
                yaml.dump(config_dict, fp, encoding=None, allow_unicode=True,
                          default_flow_style=False, indent=2,
                          Dumper=yamlloader.ordereddict.CSafeDumper)
            else:
                assert dpm_format == 'json'
                json.dump(config_dict, fp, indent=2, sort_keys=True)
    except (IOError, OSError) as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Exported DPM configuration of CPC '{c}' into DPM config "
               "file {f} in {ff} format.".
               format(c=cpc_name, f=dpm_file, ff=dpm_format.upper()))


def cmd_dpm_import(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    dpm_file = required_option(options, 'dpm_file')
    dpm_format = options['dpm_format']
    mapping_file = options['mapping_file']
    preserve_uris = options['preserve_uris']
    preserve_wwpns = options['preserve_wwpns']

    try:
        with io.open(dpm_file, 'r', encoding='utf-8') as fp:
            if dpm_format == 'yaml':
                try:
                    config_dict = yaml.safe_load(fp)
                except (yaml.parser.ParserError, yaml.scanner.ScannerError) \
                        as exc:
                    raise click_exception(
                        "Error parsing DPM configuration file {} in YAML "
                        "format: {}".format(dpm_file, exc),
                        cmd_ctx.error_format)
            else:
                assert dpm_format == 'json'
                try:
                    config_dict = json.load(fp)
                except ValueError as exc:
                    raise click_exception(
                        "Error parsing DPM configuration file {} in JSON "
                        "format: {}".format(dpm_file, exc),
                        cmd_ctx.error_format)
    except (IOError, OSError) as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if mapping_file:
        try:
            with io.open(mapping_file, 'r', encoding='utf-8') as fp:
                try:
                    mapping_obj = yaml.safe_load(fp)
                except (yaml.parser.ParserError, yaml.scanner.ScannerError) \
                        as exc:
                    raise click_exception(
                        "Error parsing adapter mapping file {} in YAML "
                        "format: {}".format(dpm_file, exc),
                        cmd_ctx.error_format)
        except (IOError, OSError) as exc:
            raise click_exception(exc, cmd_ctx.error_format)
    else:
        mapping_obj = None

    config_dict['preserve-uris'] = bool(preserve_uris)
    config_dict['preserve-wwpns'] = bool(preserve_wwpns)
    if mapping_obj:
        config_dict['adapter-mapping'] = \
            convert_adapter_mapping(cmd_ctx, mapping_obj)
    else:
        if 'adapter-mapping' in config_dict:
            del config_dict['adapter-mapping']

    result = cpc.import_dpm_configuration(config_dict)

    cmd_ctx.spinner.stop()

    if result:
        click.echo("Partially imported DPM configuration from DPM config file "
                   "{f} into CPC '{c}'. The following parts were not restored:".
                   format(c=cpc_name, f=dpm_file))
        try:
            print_dicts(cmd_ctx, result['output'], cmd_ctx.output_format)
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
    else:
        click.echo("Imported DPM configuration from DPM config file {f} into "
                   "CPC '{c}'.".
                   format(c=cpc_name, f=dpm_file))


def convert_adapter_mapping(cmd_ctx, mapping_obj):
    """
    Convert the adapter mapping from the format used in the adapter mapping file
    to the format needed by the adapter-mapping property in the Import DPM
    Configuration operation.
    """
    try:
        validate(mapping_obj, MAPPING_SCHEMA, "adapter mapping file")
    except ValueError as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    ret_mapping = []
    mapping_list = mapping_obj['adapter-mapping']
    for mapping in mapping_list:
        ret_item = {}
        ret_item['new-adapter-id'] = mapping['new-adapter-id']
        ret_item['old-adapter-id'] = mapping['old-adapter-id']
        ret_mapping.append(ret_item)
    return ret_mapping


def get_auto_start_list(cpc):
    """
    Helper functoin that converts the 'auto-start-list' property of a CPC
    to a list suitable for the zhmcclient.Cpc.set_auto_start_list() method.

    Returns:
        None - if the CPC is in classic mode
        list, with items that are one of:
          - tuple(partition, post_start_delay)
          - tuple(partition_list, name, description, post_start_delay)
    """
    auto_start_list = cpc.prop('auto-start-list', None)
    if auto_start_list is None:
        # CPC is in classic mode
        return None

    as_list = []
    for auto_start_item in auto_start_list:
        if auto_start_item['type'] == 'partition':
            # item is a partition
            uri = auto_start_item['partition-uri']
            delay = auto_start_item['post-start-delay']
            partition = cpc.partitions.resource_object(uri)
            as_item = (partition, delay)
            as_list.append(as_item)
        if auto_start_item['type'] == 'partition-group':
            # item is a partition group
            name = auto_start_item['name']
            description = auto_start_item['description']
            delay = auto_start_item['post-start-delay']
            partitions = []
            for uri in auto_start_item['partition-uris']:
                partition = cpc.partitions.resource_object(uri)
                partitions.append(partition)
            as_item = (partitions, name, description, delay)
            as_list.append(as_item)
    return as_list


def auto_start_table_str(as_list, output_format):
    """
    Return a string with the auto-start list table in the specified output
    format.
    """
    headers = ['Partition/Group', 'Post start delay', 'Partitions in group',
               'Group description']
    table = []
    for as_item in as_list:
        if isinstance(as_item[0], zhmcclient.Partition):
            # item is a partition
            partition, delay = as_item
            row = [partition.name, delay]
            table.append(row)
        else:
            # item is a partition group
            partitions, name, description, delay = as_item
            partition_names = ', '.join([p.name for p in partitions])
            row = [name, delay, partition_names, description]
            table.append(row)
    table_str = tabulate(table, headers, tablefmt=output_format)
    return table_str


def cmd_cpc_autostart_show(cmd_ctx, cpc_name):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    as_list = get_auto_start_list(cpc)

    if as_list is None:
        # The HMC WS API book documents that only CPCs in DPM mode have
        # the 'auto-start-list' property.
        cmd_ctx.spinner.stop()
        click.echo("CPC '{c}' is in classic mode and has no auto-start list.".
                   format(c=cpc_name))
        return

    table_str = auto_start_table_str(as_list, cmd_ctx.output_format)
    cmd_ctx.spinner.stop()
    click.echo(table_str)


def cmd_cpc_autostart_add(cmd_ctx, cpc_name, partitions_delay, options):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    partition_names, delay = partitions_delay

    # pylint: disable=import-outside-toplevel,cyclic-import
    from ._cmd_partition import find_partition

    as_list = get_auto_start_list(cpc)

    if as_list is None:
        # The HMC WS API book documents that only CPCs in DPM mode have
        # the 'auto-start-list' property.
        cmd_ctx.spinner.stop()
        click.echo("CPC '{c}' is in classic mode and has no auto-start list.".
                   format(c=cpc_name))
        return

    group_name = options['group']
    if group_name:
        # A partition group is added (with one or more partitions)
        partition_names = partition_names.split(',')
        partitions = []
        for partition_name in partition_names:
            partition = find_partition(cmd_ctx, client, cpc, partition_name)
            partitions.append(partition)
        description = options['description']
        new_as_item = (partitions, group_name, description, delay)
    else:
        # A partition is added
        partition_name = partition_names
        partition = find_partition(cmd_ctx, client, cpc, partition_name)
        new_as_item = (partition, delay)

    # TODO: Add support for --before and --after
    as_list.append(new_as_item)
    try:
        cpc.set_auto_start_list(as_list)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    table_str = auto_start_table_str(as_list, cmd_ctx.output_format)
    cmd_ctx.spinner.stop()
    click.echo(table_str)


def cmd_cpc_autostart_delete(cmd_ctx, cpc_name, partition_or_group):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    as_list = get_auto_start_list(cpc)

    if as_list is None:
        # The HMC WS API book documents that only CPCs in DPM mode have
        # the 'auto-start-list' property.
        cmd_ctx.spinner.stop()
        click.echo("CPC '{c}' is in classic mode and has no auto-start list.".
                   format(c=cpc_name))
        return

    as_item_idx = None
    for i, as_item in enumerate(as_list):
        if isinstance(as_item[0], zhmcclient.Partition):
            # item is a partition
            partition = as_item[0]
            if partition.name == partition_or_group:
                as_item_idx = i
                break
        else:
            # item is a partition group
            name = as_item[1]
            if name == partition_or_group:
                as_item_idx = i
                break

    if as_item_idx is None:
        raise click_exception(
            "Could not find partition or group '{p}' in CPC '{c}'.'".
            format(p=partition_or_group, c=cpc_name),
            cmd_ctx.error_format)

    del as_list[as_item_idx]
    try:
        cpc.set_auto_start_list(as_list)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    table_str = auto_start_table_str(as_list, cmd_ctx.output_format)
    cmd_ctx.spinner.stop()
    click.echo(table_str)


def cmd_cpc_autostart_clear(cmd_ctx, cpc_name):
    # pylint: disable=missing-function-docstring
    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    as_list = []
    try:
        cpc.set_auto_start_list(as_list)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Auto-start list for CPC '{c}' has been cleared.".
               format(c=cpc_name))


def cmd_cpc_list_api_features(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    name = options['name']
    try:
        features = cpc.list_api_features(name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    print_list(cmd_ctx, features, cmd_ctx.output_format)


def cmd_cpc_upgrade(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)
    console = client.consoles.console

    bundle_level = options['bundle_level']
    accept_firmware = options['accept_firmware']
    timeout = options['timeout']

    ec_mcl = console.prop('ec-mcl-description')
    hmc_bundle_level = ec_mcl.get('bundle-level', None)
    if hmc_bundle_level is None:
        hmc_version = console.prop('version')
        raise click_exception(
            "HMC version {v} does not support firmware upgrade through "
            "the Web Services API".format(v=hmc_version),
            cmd_ctx.error_format)

    click.echo("Upgrading the SE of CPC {c} to bundle level {bl} and waiting "
               "for completion (timeout: {t} s)".
               format(c=cpc_name, bl=bundle_level, t=timeout))

    try:
        cpc.single_step_install(
            bundle_level=bundle_level,
            accept_firmware=accept_firmware,
            wait_for_completion=True,
            operation_timeout=timeout)
    except zhmcclient.HTTPError as exc:
        if exc.http_status == 400 and exc.reason == 356:
            # HMC was already at that bundle level
            cmd_ctx.spinner.stop()
            click.echo("The SE of CPC {c} was already at bundle level {bl} "
                       "and did not need to be changed".
                       format(c=cpc_name, bl=bundle_level))
            return
        raise click_exception(exc, cmd_ctx.error_format)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("The SE of CPC {c} has been upgraded to bundle level {bl}".
               format(c=cpc_name, bl=bundle_level))



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for capacity groups.
"""

from __future__ import absolute_import

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS
from ._cmd_cpc import find_cpc
from ._cmd_partition import find_partition


# Default capping values for create
DEFAULT_CP_CAP = 0.0
DEFAULT_IFL_CAP = 0.0

# Limits for capping values
MIN_CAP = 0.0
MAX_CAP = 255.0


def find_capacitygroup(cmd_ctx, client, cpc_or_name, capacitygroup_name):
    """
    Find a capacity group by name and return its resource object.
    """
    if isinstance(cpc_or_name, zhmcclient.Cpc):
        cpc = cpc_or_name
    else:
        cpc = find_cpc(cmd_ctx, client, cpc_or_name)
    # The CPC must be in DPM mode. We don't check that because it would
    # cause a GET to the CPC resource that we otherwise don't need.
    try:
        capacitygroup = cpc.capacity_groups.find(name=capacitygroup_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return capacitygroup


@cli.group('capacitygroup', options_metavar=COMMAND_OPTIONS_METAVAR)
def capacitygroup_group():
    """
    Command group for managing capacity groups.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@capacitygroup_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@add_options(LIST_OPTIONS)
@click.pass_obj
def capacitygroup_list(cmd_ctx, cpc, **options):
    """
    List the capacity groups in a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_capacitygroup_list(cmd_ctx, cpc, options))


@capacitygroup_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('CAPACITYGROUP', type=str, metavar='CAPACITYGROUP')
@click.pass_obj
def capacitygroup_show(cmd_ctx, cpc, capacitygroup):
    """
    Show the details of a capacity group.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent CPC.
      - 'partition-names' - Names of the partitions referenced by
        'partition-uris' (index-correlated).

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_capacitygroup_show(cmd_ctx, cpc, capacitygroup))


@capacitygroup_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('CAPACITYGROUP', type=str, metavar='CAPACITYGROUP')
@click.option('--name', type=str, required=False,
              help='The new name of the capacity group.')
@click.option('--description', type=str, required=False,
              help='The new description of the capacity group.')
@click.option('--enabled/--no-enabled', is_flag=True, required=False,
              default=None,
              help='The new enabled state of the capacity group.')
@click.option('--cp-cap', type=float, required=False, default=None,
              help='The new absolute capping value for CP processors, in '
              'units of processors from {mn} to {mx}.'.
              format(mn=MIN_CAP, mx=MAX_CAP))
@click.option('--ifl-cap', type=float, required=False, default=None,
              help='The new absolute capping value for IFL processors, in '
              'units of processors from {mn} to {mx}.'.
              format(mn=MIN_CAP, mx=MAX_CAP))
@click.pass_obj
def capacitygroup_update(cmd_ctx, cpc, capacitygroup, **options):
    """
    Update the properties of a capacity group.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_capacitygroup_update(cmd_ctx, cpc, capacitygroup, options))


@capacitygroup_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.option('--name', type=str, required=True,
              help='The name of the new capacity group.')
@click.option('--description', type=str, required=False,
              help='The description of the new capacity group. Default: empty')
@click.option('--enabled/--no-enabled', is_flag=True, required=False,
              default=True,
              help='The enabled state of the new capacity group. '
              'Default: Enabled')
@click.option('--cp-cap', type=float, required=False, default=None,
              help='The absolute capping value for CP processors, in units of '
              'processors from {mn} to {mx}. Default: {df}'.
              format(mn=MIN_CAP, mx=MAX_CAP, df=DEFAULT_CP_CAP))
@click.option('--ifl-cap', type=float, required=False, default=None,
              help='The absolute capping value for IFL processors, in units of '
              'processors from {mn} to {mx}. Default: {df}'.
              format(mn=MIN_CAP, mx=MAX_CAP, df=DEFAULT_IFL_CAP))
@click.pass_obj
def capacitygroup_create(cmd_ctx, cpc, **options):
    """
    Create a capacity group in a CPC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_capacitygroup_create(cmd_ctx, cpc, options))


@capacitygroup_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('CAPACITYGROUP', type=str, metavar='CAPACITYGROUP')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the capacity group.',
              prompt='Are you sure you want to delete this capacity group ?')
@click.pass_obj
def capacitygroup_delete(cmd_ctx, cpc, capacitygroup):
    """
    Delete a capacity group.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_capacitygroup_delete(cmd_ctx, cpc, capacitygroup))


@capacitygroup_group.command(
    'add-partition', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('CAPACITYGROUP', type=str, metavar='CAPACITYGROUP')
@click.option('--partition', type=str, metavar='NAME',
              required=False, multiple=True,
              help='The name of the partition to be added to the capacity '
              'group. Can be specified multiple times.')
@click.pass_obj
def capacitygroup_add_partition(cmd_ctx, cpc, capacitygroup, **options):
    """
    Add a partition to a capacity group.

    A partition cannot become a member of more than one capacity group.
    The partition must be defined with shared processors and the
    capacity group must specify nonzero cap values for the processor types
    used by the partition. The partition must be on the same CPC as the
    capacity group and must not yet be a member of this (or any other)
    capacity group.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_capacitygroup_add_partition(
            cmd_ctx, cpc, capacitygroup, options))


@capacitygroup_group.command(
    'remove-partition', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('CPC', type=str, metavar='CPC')
@click.argument('CAPACITYGROUP', type=str, metavar='CAPACITYGROUP')
@click.option('--partition', type=str, metavar='NAME',
              required=False, multiple=True,
              help='The name of the partition to be removed from the capacity '
              'group. Can be specified multiple times.')
@click.pass_obj
def capacitygroup_remove_partition(cmd_ctx, cpc, capacitygroup, **options):
    """
    Remove a partition from a capacity group.

    The partition must be a member of this capacity group.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(
        lambda: cmd_capacitygroup_remove_partition(
            cmd_ctx, cpc, capacitygroup, options))


def cmd_capacitygroup_list(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    try:
        capacitygroups = cpc.capacity_groups.list()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    show_list = [
        'name',
        'cpc',
    ]
    if not options['names_only']:
        show_list.extend([
            'capping-enabled',
            'absolute-general-purpose-proc-cap',
            'absolute-ifl-proc-cap',
            'partitions',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'element-uri',
        ])

    cpc_additions = {}
    partitions_additions = {}
    for capacitygroup in capacitygroups:
        cpc_additions[capacitygroup.uri] = cpc_name
        try:
            partition_uris = capacitygroup.prop('partition-uris')
            partition_names = []
            for partition_uri in partition_uris:
                partition = cpc.partitions.find(**{'object-uri': partition_uri})
                partition_names.append(partition.name)
            partitions_additions[capacitygroup.uri] = partition_names
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
    additions = {
        'cpc': cpc_additions,
        'partitions': partitions_additions,
    }

    try:
        print_resources(cmd_ctx, capacitygroups, cmd_ctx.output_format,
                        show_list, additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_capacitygroup_show(cmd_ctx, cpc_name, capacitygroup_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    capacitygroup = find_capacitygroup(
        cmd_ctx, client, cpc_name, capacitygroup_name)

    try:
        capacitygroup.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(capacitygroup.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = cpc_name

    # Add artificial property 'partition-names'
    part_names = []
    for part_uri in capacitygroup.properties['partition-uris']:
        part_props = client.session.get(part_uri)
        part_names.append(part_props['name'])
    properties['partition-names'] = part_names

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_capacitygroup_update(cmd_ctx, cpc_name, capacitygroup_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    capacitygroup = find_capacitygroup(
        cmd_ctx, client, cpc_name, capacitygroup_name)

    name_map = {
        'enabled': 'capping-enabled',
        'cp-cap': 'absolute-general-purpose-proc-cap',
        'ifl-cap': 'absolute-ifl-proc-cap',
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating "
                   "capacity group '{cg}'.".
                   format(cg=capacitygroup_name))
        return

    try:
        capacitygroup.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    if 'name' in properties and properties['name'] != capacitygroup_name:
        click.echo("Capacity group '{cg}' has been renamed to '{cgn}' and was "
                   "updated.".
                   format(cg=capacitygroup_name, cgn=properties['name']))
    else:
        click.echo("Capacity group '{cg}' has been updated.".
                   format(cg=capacitygroup_name))


def cmd_capacitygroup_create(cmd_ctx, cpc_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    name_map = {
        'enabled': 'capping-enabled',
        'cp-cap': 'absolute-general-purpose-proc-cap',
        'ifl-cap': 'absolute-ifl-proc-cap',
    }
    org_options = original_options(options)
    properties = options_to_properties(org_options, name_map)

    try:
        new_capacitygroup = cpc.capacity_groups.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New capacity group '{cg}' has been created.".
               format(cg=new_capacitygroup.properties['name']))


def cmd_capacitygroup_delete(cmd_ctx, cpc_name, capacitygroup_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    capacitygroup = find_capacitygroup(
        cmd_ctx, client, cpc_name, capacitygroup_name)

    try:
        capacitygroup.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("Capacity group '{cg}' has been deleted.".
               format(cg=capacitygroup_name))


def cmd_capacitygroup_add_partition(
        cmd_ctx, cpc_name, capacitygroup_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    capacitygroup = find_capacitygroup(
        cmd_ctx, client, cpc_name, capacitygroup_name)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    partition_names = options['partition']  # List
    if not partition_names:
        cmd_ctx.spinner.stop()
        click.echo("No partitions specified for adding to "
                   "capacity group '{cg}'.".
                   format(cg=capacitygroup_name))
        return

    for partition_name in partition_names:
        partition = find_partition(cmd_ctx, client, cpc, partition_name)
        try:
            capacitygroup.add_partition(partition)
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("The specified partitions have been added to "
               "capacity group '{cg}'.".
               format(cg=capacitygroup_name))


def cmd_capacitygroup_remove_partition(
        cmd_ctx, cpc_name, capacitygroup_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    capacitygroup = find_capacitygroup(
        cmd_ctx, client, cpc_name, capacitygroup_name)
    cpc = find_cpc(cmd_ctx, client, cpc_name)

    partition_names = options['partition']  # List
    if not partition_names:
        cmd_ctx.spinner.stop()
        click.echo("No partitions specified for removing from "
                   "capacity group '{cg}'.".
                   format(cg=capacitygroup_name))
        return

    for partition_name in partition_names:
        partition = find_partition(cmd_ctx, client, cpc, partition_name)
        try:
            capacitygroup.remove_partition(partition)
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("The specified partitions have been removed from "
               "capacity group '{cg}'.".
               format(cg=capacitygroup_name))



# Copyright 2016-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main script.
"""

from __future__ import absolute_import, print_function

import os
import sys
import logging
from logging.handlers import SysLogHandler
from logging import StreamHandler, NullHandler
import platform
import urllib3
import click
import click_repl
from prompt_toolkit.history import FileHistory

import zhmcclient
import zhmcclient_mock
from ._helper import CmdContext, GENERAL_OPTIONS_METAVAR, REPL_HISTORY_FILE, \
    REPL_PROMPT, TABLE_FORMATS, LOG_LEVELS, LOG_DESTINATIONS, \
    SYSLOG_FACILITIES, click_exception, get_click_terminal_width

urllib3.disable_warnings()


# Default values for some options
DEFAULT_OUTPUT_FORMAT = 'table'
DEFAULT_ERROR_FORMAT = 'msg'
DEFAULT_TIMESTATS = False
DEFAULT_LOG = 'all=warning'
DEFAULT_LOG_DESTINATION = 'stderr'
DEFAULT_SYSLOG_FACILITY = 'user'
DEFAULT_NO_VERIFY = False

CONSOLE_LOGGER_NAME = 'zhmccli.console'

ERROR_FORMATS = ['msg', 'def']

# List of values to try for the 'address' parameter when creating
# a SysLogHandler object.
# Key: Operating system type, as returned by platform.system(). For CygWin,
# the returned value is 'CYGWIN_NT-6.1', which is special-cased to 'CYGWIN_NT'.
# Value: List of values for the 'address' parameter; to be tried in the
# specified order.
SYSLOG_ADDRESSES = {
    'Linux': ['/dev/log', ('localhost', 514)],
    'Darwin': ['/var/run/syslog', ('localhost', 514)],  # OS-X
    'Windows': [('localhost', 514)],
    'CYGWIN_NT': ['/dev/log', ('localhost', 514)],  # Requires syslog-ng pkg
}

# The getattr() is used to work around a Pylint false positive no-member issue
ZHMCCLIENT_VERSION = "zhmcclient, version {v}".format(
    v=getattr(zhmcclient, '__version__', 'unknown'))

# Logger names by log component
LOGGER_NAMES = {
    'all': '',  # root logger
    'api': zhmcclient.API_LOGGER_NAME,
    'hmc': zhmcclient.HMC_LOGGER_NAME,
    'console': CONSOLE_LOGGER_NAME,
}
LOG_COMPONENTS = LOGGER_NAMES.keys()

# Context variables passed to Click
CLICK_CONTEXT_SETTINGS = dict(

    # Set the terminal width - used e.g. for Click help messages
    terminal_width=get_click_terminal_width(),
)


@click.group(invoke_without_command=True,
             context_settings=CLICK_CONTEXT_SETTINGS,
             options_metavar=GENERAL_OPTIONS_METAVAR)
@click.option('-h', '--host', type=str, envvar='ZHMC_HOST',
              help="Hostname or IP address of the HMC "
                   "(Default: ZHMC_HOST environment variable).")
@click.option('-u', '--userid', type=str, envvar='ZHMC_USERID',
              help="Username for the HMC "
                   "(Default: ZHMC_USERID environment variable).")
@click.option('-p', '--password', type=str, envvar='ZHMC_PASSWORD',
              help="Password for the HMC "
                   "(Default: ZHMC_PASSWORD environment variable).")
@click.option('-n', '--no-verify', is_flag=True, default=None,
              envvar='ZHMC_NO_VERIFY',
              help="Do not verify the HMC certificate. "
                   "(Default: ZHMC_NO_VERIFY environment variable, or verify "
                   "the HMC certificate).")
@click.option('-c', '--ca-certs', type=str, envvar='ZHMC_CA_CERTS',
              help="Path name of certificate file or directory with CA "
                   "certificates to be used for verifying the HMC certificate. "
                   "(Default: Path name in ZHMC_CA_CERTS environment variable, "
                   "or path name in REQUESTS_CA_BUNDLE environment variable, "
                   "or path name in CURL_CA_BUNDLE environment variable, "
                   "or the 'certifi' Python package which provides the "
                   "Mozilla CA Certificate List).")
@click.option('-o', '--output-format',
              type=click.Choice(TABLE_FORMATS + ['json']),
              help='Output format (Default: {def_of}).'.
              format(def_of=DEFAULT_OUTPUT_FORMAT))
@click.option('-x', '--transpose', type=str, is_flag=True,
              help='Transpose the output table for metrics.')
@click.option('-e', '--error-format', type=click.Choice(ERROR_FORMATS),
              help='Error message format (Default: {def_ef}).'.
              format(def_ef=DEFAULT_ERROR_FORMAT))
@click.option('-t', '--timestats', type=str, is_flag=True,
              help='Show time statistics of HMC operations.')
@click.option('--log', type=str, metavar='COMP=LEVEL,...',
              help="Set a component to a log level (COMP: [{comps}], "
              "LEVEL: [{levels}], Default: {def_log}).".
              format(comps='|'.join(LOG_COMPONENTS),
                     levels='|'.join(LOG_LEVELS),
                     def_log=DEFAULT_LOG))
@click.option('--log-dest', type=str,
              metavar='[{}]'.format('|'.join(LOG_DESTINATIONS)),
              help="Log destination for this command (Default: {def_dest}).".
              format(def_dest=DEFAULT_LOG_DESTINATION))
@click.option('--syslog-facility', type=click.Choice(SYSLOG_FACILITIES),
              help="Syslog facility when logging to the syslog "
              "(Default: {def_slf}).".
              format(def_slf=DEFAULT_SYSLOG_FACILITY))
@click.option('--pdb', is_flag=True, hidden=True,
              help=u'Break execution in the pdb debugger just before '
                   u'executing the command within zhmc.')
@click.version_option(
    message='%(prog)s, version %(version)s\n' + ZHMCCLIENT_VERSION,
    help="Show the versions of this command and of the zhmcclient package and "
    "exit.")
@click.pass_context
def cli(ctx, host, userid, password, no_verify, ca_certs, output_format,
        transpose, error_format, timestats, log, log_dest, syslog_facility,
        pdb):
    """
    Command line interface for the IBM Z HMC.

    The options shown in this help text are general options that can also
    be specified on any of the (sub-)commands.
    """

    # Concept: In interactive mode, the global options specified in the command
    # line are used as defaults for the commands that are issued interactively.
    # The interactive commands may override these options.
    # This requires being able to determine for each option whether it has been
    # specified. This is the reason the options don't define defaults in the
    # decorators that define them.

    if ctx.obj is None:
        # We are in command mode or are processing the command line options in
        # interactive mode.
        # We apply the documented option defaults.
        if output_format is None:
            output_format = DEFAULT_OUTPUT_FORMAT
        if transpose is None:
            transpose = False
        if error_format is None:
            error_format = DEFAULT_ERROR_FORMAT
        if timestats is None:
            timestats = DEFAULT_TIMESTATS
        if no_verify is None:
            no_verify = DEFAULT_NO_VERIFY
    else:
        # We are processing an interactive command.
        # We apply the option defaults from the command line options.
        if host is None:
            host = ctx.obj.host
        if userid is None:
            userid = ctx.obj.userid
        if password is None:
            # pylint: disable=protected-access
            password = ctx.obj._password
        if no_verify is None:
            no_verify = ctx.obj.no_verify
        if ca_certs is None:
            ca_certs = ctx.obj.ca_certs
        if output_format is None:
            output_format = ctx.obj.output_format
        if transpose is None:
            transpose = ctx.obj.transpose
        if error_format is None:
            error_format = ctx.obj.error_format
        if timestats is None:
            timestats = ctx.obj.timestats
        if pdb is None:
            pdb = ctx.obj.pdb

    if transpose and output_format == 'json':
        raise click_exception(
            "Transposing output tables (-x / --transpose) conflicts with "
            "non-table output format (-o / --output-format): {of}".
            format(of=output_format),
            error_format)

    if no_verify and ca_certs:
        raise click_exception(
            "Disabling HMC certificate verification (-n / --no-verify / "
            "ZHMC_NO_VERIFY) conflicts with specifying a CA certificate path "
            "(-c / --ca-certs / ZHMC_CA_CERTS)",
            error_format)

    # TODO: Add context support for the following options:
    if log is None:
        log = DEFAULT_LOG
    if log_dest is None:
        log_dest = DEFAULT_LOG_DESTINATION
    if syslog_facility is None:
        syslog_facility = DEFAULT_SYSLOG_FACILITY

    # Now we have the effective values for the options as they should be used
    # by the current command, regardless of the mode.

    # Set up logging

    for lc in LOG_COMPONENTS:
        reset_logger(lc)

    if log_dest == 'none':
        handler = None
    elif log_dest == 'syslog':
        # The choices in SYSLOG_FACILITIES have been validated by click
        # so we don't need to further check them.
        facility = SysLogHandler.facility_names[syslog_facility]
        system = platform.system()
        if system.startswith('CYGWIN_NT'):
            # Value is 'CYGWIN_NT-6.1'; strip off trailing version:
            system = 'CYGWIN_NT'
        try:
            addresses = SYSLOG_ADDRESSES[system]
        except KeyError:
            raise NotImplementedError(
                "Logging to syslog is not supported on this platform: {p}".
                format(p=system))
        assert isinstance(addresses, list)
        for address in addresses:
            try:
                handler = SysLogHandler(address=address, facility=facility)
            except Exception:  # pylint: disable=broad-except
                continue
            break
        else:
            exc = sys.exc_info()[1]
            exc_name = exc.__class__.__name__ if exc else None
            raise RuntimeError(
                "Creating SysLogHandler with addresses {al!r} failed. "
                "Failure on last address {a!r} was: {exc}: {msg}".
                format(al=addresses, a=address, exc=exc_name, msg=exc))
        fs = '%(levelname)s %(name)s: %(message)s'
        handler.setFormatter(logging.Formatter(fs))
    elif log_dest == 'stderr':
        handler = StreamHandler(stream=sys.stderr)
        fs = '%(levelname)s %(name)s: %(message)s'
        handler.setFormatter(logging.Formatter(fs))
    else:
        # log_dest is the path name of the log file
        try:
            handler = logging.FileHandler(log_dest)
        except OSError as exc:
            raise click_exception(
                "Cannot log to file {fn}: {exc}: {msg}".
                format(fn=log_dest, exc=exc.__class__.__name__, msg=exc),
                error_format)
        fs = '%(levelname)s %(name)s: %(message)s'
        handler.setFormatter(logging.Formatter(fs))

    log_specs = log.split(',')
    for log_spec in log_specs:

        # ignore extra ',' at begin, end or in between
        if log_spec == '':
            continue

        try:
            log_comp, log_level = log_spec.split('=', 1)
        except ValueError:
            raise click_exception("Missing '=' in COMP=LEVEL specification in "
                                  "--log option: {ls}".format(ls=log_spec),
                                  error_format)

        level = getattr(logging, log_level.upper(), None)
        if level is None:
            raise click_exception("Invalid log level in COMP=LEVEL "
                                  "specification in --log option: {ls}".
                                  format(ls=log_spec),
                                  error_format)

        if log_comp not in LOG_COMPONENTS:
            raise click_exception("Invalid log component in COMP=LEVEL "
                                  "specification in --log option: {ls}".
                                  format(ls=log_spec), error_format)

        if handler:
            setup_logger(log_comp, handler, level)

    session_id = os.environ.get('ZHMC_SESSION_ID', None)
    if session_id and session_id.startswith('faked_session:'):
        # This should be used by the zhmc function tests only.
        # A SyntaxError raised by an incorrect expression is considered
        # an internal error in the function tests and is therefore not
        # handled.
        expr = session_id.split(':', 1)[1]
        faked_session = eval(expr)  # pylint: disable=eval-used
        assert isinstance(faked_session, zhmcclient_mock.FakedSession)
        session_id = faked_session

    def get_password_via_prompt(host, userid):
        """
        Password retrieval function that prompts for the password.

        It follows the interface defined in
        :func:`~zhmcclient.get_password_interface` and needs access to the
        click context (ctx).
        """
        if userid is not None and host is not None:
            ctx.obj.spinner.stop()
            password = click.prompt(
                "Enter password (for user {userid} at HMC {host})".
                format(userid=userid, host=host), hide_input=True,
                confirmation_prompt=False, type=str, err=True)
            ctx.obj.spinner.start()
            return password

        raise click_exception("{cmd} command requires logon, but no "
                              "session-id or userid provided.".
                              format(cmd=ctx.invoked_subcommand),
                              error_format)

    # We create a command context for each command: An interactive command has
    # its own command context different from the command context for the
    # command line.
    ctx.obj = CmdContext(host, userid, password, no_verify, ca_certs,
                         output_format, transpose, error_format, timestats,
                         session_id, get_password_via_prompt, pdb)

    # Invoke default command
    if ctx.invoked_subcommand is None:
        ctx.invoke(repl)


def reset_logger(log_comp):
    """
    Reset the logger for the specified log component to have exactly one
    NullHandler.
    """

    name = LOGGER_NAMES[log_comp]
    logger = logging.getLogger(name)

    has_nh = False
    for h in logger.handlers:
        if not has_nh and isinstance(h, NullHandler):
            has_nh = True
            continue
        logger.removeHandler(h)

    if not has_nh:
        nh = NullHandler()
        logger.addHandler(nh)


def setup_logger(log_comp, handler, level):
    """
    Setup the logger for the specified log component to add the specified
    handler and to set it to the specified log level.
    """

    name = LOGGER_NAMES[log_comp]
    logger = logging.getLogger(name)

    logger.addHandler(handler)
    logger.setLevel(level)


@cli.command('help')
@click.pass_context
def repl_help(ctx):
    # pylint: disable=unused-argument
    """
    Show help message for interactive mode.

    Parameters:

      ctx (:class:`click.Context`): The click context object. Created by the
        ``@click.pass_context`` decorator.
    """
    print("""
The following can be entered in interactive mode:

  <zhmc-cmd>                  Execute zhmc command <zhmc-cmd>.
  !<shell-cmd>                Execute shell command <shell-cmd>.

  <CTRL-D>, :q, :quit, :exit  Exit interactive mode.

  <TAB>                       Tab completion (can be used anywhere).
  --help                      Show zhmc general help message, including a list
                              of zhmc commands.
  <zhmc-cmd> --help           Show help message for zhmc command <zhmc-cmd>.
  help                        Show this help message.
  :?, :h, :help               Show (incomplete) help message about interactive
                              mode.
""")


@cli.command('repl')
@click.pass_context
def repl(ctx):
    """
    Enter interactive (REPL) mode (default).

    Parameters:

      ctx (:class:`click.Context`): The click context object. Created by the
        ``@click.pass_context`` decorator.
    """

    history_file = REPL_HISTORY_FILE
    if history_file.startswith('~'):
        history_file = os.path.expanduser(history_file)

    print("Enter 'help' for help, <CTRL-D> or ':q' to exit.")

    prompt_kwargs = {
        'message': REPL_PROMPT,
        'history': FileHistory(history_file),
    }
    click_repl.repl(ctx, prompt_kwargs=prompt_kwargs)


# TODO: Apparently registering is not needed, clarify that.
# click_repl.register_repl(repl)



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for HMC user role management.
"""

from __future__ import absolute_import
from __future__ import print_function

import re

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, print_resources, abort_if_false, \
    options_to_properties, original_options, COMMAND_OPTIONS_METAVAR, \
    click_exception, add_options, LIST_OPTIONS, ObjectByUriCache


PERMISSION_OPTIONS = [
    click.option('--task', type=str, metavar='TASK', required=False,
                 help='Permission: Task permission to the task with the '
                 'specified name.'),
    click.option('--view-only', type=bool, required=False,
                 help='If --task was specified, indicates whether the task\'s '
                 'view-only version is subject of the permission. Only certain '
                 'tasks have a view-only version. Default: True.'),
    click.option('--class', type=str, metavar='CLASS', required=False,
                 help='Permission: Object permission to all objects of the '
                 'specified resource class (= value of "class" property).'),
    click.option('--group', type=str, metavar='GROUP', required=False,
                 help='Permission: Object permission to the group with the '
                 'specified name and optionally to its members.'),
    click.option('--include-members', type=bool, required=False,
                 help='If --group was specified, indicates whether the group '
                 'members are included in the permission. Default: False.'),
    click.option('--cpc', type=str, metavar='CPC', required=False,
                 help='Permission: Object permission to the CPC with the '
                 'specified name.'),
    click.option('--partition', type=str,
                 metavar='CPC PARTITION', nargs=2, required=False,
                 help='Permission: Object permission to the partition on the '
                 'CPC with the specified names.'),
    click.option('--lpar', type=str,
                 metavar='CPC LPAR', nargs=2, required=False,
                 help='Permission: Object permission to the LPAR on the CPC '
                 'with the specified names.'),
    click.option('--adapter', type=str,
                 metavar='CPC ADAPTER', nargs=2, required=False,
                 help='Permission: Object permission to the adapter on the CPC '
                 'with the specified names.'),
    click.option('--storage-group', type=str,
                 metavar='CPC STORAGEGROUP', nargs=2, required=False,
                 help='Permission: Object permission to the storage group '
                 'associated with the CPC with the specified names.'),
    click.option('--storage-group-template', type=str,
                 metavar='CPC STORAGEGROUPTEMPLATE', nargs=2, required=False,
                 help='Permission: Object permission to the storage group '
                 'template associated with the CPC with the specified names.'),
]


def find_user_role(cmd_ctx, console, user_role_name):
    """
    Find a user role by name and return its resource object.
    """
    try:
        user_role = console.user_roles.find(name=user_role_name)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)
    return user_role


def permission_str(obj_cache, permission_info):
    """
    Return a permission-info item as a string for displaying
    """
    obj = permission_info['permitted-object']
    obj_type = permission_info['permitted-object-type']
    if obj_type == 'object-class':
        perm_str = 'class {}'.format(obj)
    else:
        assert obj_type == 'object'
        if obj.startswith('/api/console/tasks/'):
            task = obj_cache.task_by_uri(obj)
            perm_str = "task '{}'".format(task.name)
        elif obj.startswith('/api/groups/'):
            # TODO: Display name, once zhmcclient.Group implemented
            # group = obj_cache.group_by_uri(obj)
            # perm_str = "group '{}'".format(group.name)
            perm_str = "group '{}'".format(obj)
        elif re.match(r'^/api/cpcs/[^/]+/adapters/[^/]+$', obj):
            adapter = obj_cache.adapter_by_uri(obj)
            cpc = adapter.manager.parent
            perm_str = "adapter '{}' on cpc '{}'". \
                format(adapter.name, cpc.name)
        elif re.match(r'^/api/cpcs/[^/]+/lpars/[^/]+$', obj):
            lpar = obj_cache.lpar_by_uri(obj)
            cpc = lpar.manager.parent
            perm_str = "lpar '{}' on cpc '{}'". \
                format(lpar.name, cpc.name)
        elif re.match(r'^/api/cpcs/[^/]+/partitions/[^/]+$', obj):
            partition = obj_cache.partition_by_uri(obj)
            cpc = partition.manager.parent
            perm_str = "partition '{}' on cpc '{}'". \
                format(partition.name, cpc.name)
        elif re.match(r'^/api/cpcs/[^/]+$', obj):
            cpc = obj_cache.cpc_by_uri(obj)
            perm_str = "cpc '{}'".format(cpc.name)
        elif re.match(r'^/api/storage-groups/[^/]+$', obj):
            storage_group = obj_cache.storage_group_by_uri(obj)
            cpc = storage_group.cpc
            perm_str = "storage group '{}' associated with cpc '{}'". \
                format(storage_group.name, cpc.name)
        elif re.match(r'^/api/storage-templates/[^/]+$', obj):
            # TODO: Display name, once zhmcclient list templates implemented
            # perm_str = "group '{}'".format(obj)
            # storage_template = obj_cache.storage_template_by_uri(obj)
            # cpc = storage_template.cpc
            # perm_str = "storage group template '{}' assoc. with cpc '{}'". \
            #     format(storage_template.name, cpc.name)
            perm_str = "storage group template '{}'".format(obj)
        else:
            perm_str = 'object {}'.format(obj)
    return perm_str


def permission_options_to_kwargs(cmd_ctx, client, options):
    """
    Convert the permission options to a kwargs dict of arguments for the
    respective zhmcclient methods.
    """
    opts = dict(options)
    kwargs = {}
    kwargs['include_members'] = options.pop('include_members')
    kwargs['view_only'] = options.pop('view_only')
    num_options = \
        bool(opts['task']) + \
        bool(opts['class']) + \
        bool(opts['group']) + \
        bool(opts['cpc']) + \
        bool(opts['partition']) + \
        bool(opts['adapter']) + \
        bool(opts['storage_group']) + \
        bool(opts['storage_group_template'])

    if num_options > 1:
        raise click.ClickException(
            "More than one permission option specified: {}".
            format(', '.join(opts.keys())))

    if opts['task']:
        raise NotImplementedError  # pylint: disable=no-else-raise
        # TODO: Implement once zhmccli task is implemented
        # # pylint: disable=import-outside-toplevel,cyclic-import
        # from ._cmd_task import find_task
        # task_name = opts['task']
        # task = find_task(cmd_ctx, client, task_name)
        # kwargs['permitted_object'] = task
    if opts['class']:
        class_name = opts['class']
        kwargs['permitted_object'] = class_name
    elif opts['group']:
        raise NotImplementedError  # pylint: disable=no-else-raise
        # TODO: Implement once zhmcclient.Group is implemented
        # # pylint: disable=import-outside-toplevel,cyclic-import
        # from ._cmd_group import find_group
        # group_name = opts['group']
        # group = find_group(cmd_ctx, client, group_name)
        # kwargs['permitted_object'] = group
    elif opts['cpc']:
        # pylint: disable=import-outside-toplevel,cyclic-import
        from ._cmd_cpc import find_cpc
        cpc_name = opts['cpc']
        cpc = find_cpc(cmd_ctx, client, cpc_name)
        kwargs['permitted_object'] = cpc
    elif opts['partition']:
        # pylint: disable=import-outside-toplevel,cyclic-import
        from ._cmd_partition import find_partition
        cpc_name, partition_name = opts['partition']
        partition = find_partition(cmd_ctx, client, cpc_name, partition_name)
        kwargs['permitted_object'] = partition
    elif opts['adapter']:
        # pylint: disable=import-outside-toplevel,cyclic-import
        from ._cmd_adapter import find_adapter
        cpc_name, adapter_name = opts['adapter']
        adapter = find_adapter(cmd_ctx, client, cpc_name, adapter_name)
        kwargs['permitted_object'] = adapter
    elif opts['storage_group']:
        # pylint: disable=import-outside-toplevel,cyclic-import
        from ._cmd_storagegroup import find_storagegroup
        cpc_name, stogrp_name = opts['storage_group']
        # TODO: Fix find_storagegroup() to take cpc_name
        stogrp = find_storagegroup(cmd_ctx, client, stogrp_name)
        kwargs['permitted_object'] = stogrp
    elif opts['storage_group_template']:
        raise NotImplementedError  # pylint: disable=no-else-raise
        # TODO: Implement once zhmccli storage-group-template is implemented
        # # pylint: disable=import-outside-toplevel,cyclic-import
        # from ._cmd_storagegrouptemplate import find_storagegrouptemplate
        # cpc_name, stogrptpl_name = opts['storage_group_template']
        # stogrptpl = find_storagegrouptemplate(
        #     cmd_ctx, client, cpc_name, stogrptpl_name)
        # kwargs['permitted_object'] = stogrptpl
    else:
        raise click.ClickException("No permission option specified.")

    return kwargs


@cli.group('userrole', options_metavar=COMMAND_OPTIONS_METAVAR)
def user_role_group():
    """
    Command group for managing HMC user roles.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """


@user_role_group.command('list', options_metavar=COMMAND_OPTIONS_METAVAR)
@add_options(LIST_OPTIONS)
@click.option('--permissions', is_flag=True, required=False,
              help='Show additional properties for user role permissions.')
@click.pass_obj
def user_role_list(cmd_ctx, **options):
    """
    List the HMC user roles.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_role_list(cmd_ctx, options))


@user_role_group.command('show', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER_ROLE', type=str, metavar='USER_ROLE')
@click.pass_obj
def user_role_show(cmd_ctx, user_role):
    """
    Show the details of an HMC user role.

    The following properties are shown in addition to those returned by the HMC:

    \b
      - 'parent-name' - Name of the parent Console.
      - 'associated-system-defined-user-role-name' - Name of the User Role
        referenced by 'associated-system-defined-user-role-uri', if present.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_role_show(cmd_ctx, user_role))


@user_role_group.command('create', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.option('--name', type=str, required=True,
              help='The name of the new user role.')
@click.option('--description', type=str, required=False,
              help='The description of the new user role.')
@click.option('--associated-system-defined-user-role', type=str, required=False,
              help='The system-defined user role associated with the new '
              'user role. '
              'Default: Operator Tasks')
@click.option('--is-inheritance-enabled', type=bool, required=False,
              help='The indicator whether user role inheritance is enabled for '
              'the new user role. '
              'When enabled, if the user role permits access to a parent '
              'managed object, then all managed objects that are hosted by the '
              'parent managed object are also permitted by this role. '
              'Default: False')
@click.pass_obj
def user_role_create(cmd_ctx, **options):
    """
    Create a user-defined HMC user role.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_role_create(cmd_ctx, options))


@user_role_group.command('update', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER_ROLE', type=str, metavar='USER_ROLE')
@click.option('--description', type=str, required=False,
              help='The new description of the user role.')
@click.option('--associated-system-defined-user-role', type=str, required=False,
              help='The new system-defined user role associated with the '
              'user role.')
@click.option('--is-inheritance-enabled', type=bool, required=False,
              help='The new indicator whether user role inheritance is '
              'enabled for the user role. '
              'When enabled, if the user role permits access to a parent '
              'managed object, then all managed objects that are hosted by the '
              'parent managed object are also permitted by this role.')
@click.pass_obj
def user_role_update(cmd_ctx, user_role, **options):
    """
    Update the properties of an HMC user role.

    Note that HMC user roles cannot be renamed.

    Only the properties will be changed for which a corresponding option is
    specified, so the default for all options is not to change properties.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_role_update(
        cmd_ctx, user_role, options))


@user_role_group.command('delete', options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER_ROLE', type=str, metavar='USER_ROLE')
@click.option('-y', '--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              help='Skip prompt to confirm deletion of the user role.',
              prompt='Are you sure you want to delete this user role ?')
@click.pass_obj
def user_role_delete(cmd_ctx, user_role):
    """
    Delete a user-defined HMC user role.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_role_delete(cmd_ctx, user_role))


@user_role_group.command('add-permission',
                         options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER_ROLE', type=str, metavar='USER_ROLE')
@add_options(PERMISSION_OPTIONS)
@click.pass_obj
def user_role_add_permission(cmd_ctx, user_role, **options):
    """
    Add a permission to a user-defined HMC user role.

    The permission is specified by exactly one of the permission options,
    and optionally the --view-only or --include-members options.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_role_add_permission(
        cmd_ctx, user_role, options))


@user_role_group.command('remove-permission',
                         options_metavar=COMMAND_OPTIONS_METAVAR)
@click.argument('USER_ROLE', type=str, metavar='USER_ROLE')
@add_options(PERMISSION_OPTIONS)
@click.pass_obj
def user_role_remove_permission(cmd_ctx, user_role, **options):
    """
    Remove a permission from a user-defined HMC user role.

    The permission is specified by exactly one of the permission options,
    and optionally the --view-only or --include-members options.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified right after the
    'zhmc' command name.
    """
    cmd_ctx.execute_cmd(lambda: cmd_user_role_remove_permission(
        cmd_ctx, user_role, options))


def cmd_user_role_list(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    show_list = [
        'name',
    ]
    if not options['names_only']:
        show_list.extend([
            'type',
            'description',
        ])
    if options['uri']:
        show_list.extend([
            'object-uri',
        ])
    if options['permissions']:
        show_list.extend([
            'permissions',
            'is-inheritance-enabled',
        ])

    additions = {}
    additions['permissions'] = {}

    try:
        user_roles = console.user_roles.list(full_properties=False)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    if options['permissions']:
        obj_cache = ObjectByUriCache(cmd_ctx, client)
        for role in user_roles:
            permissions = []
            for item in role.get_property('permissions'):
                permissions.append(permission_str(obj_cache, item))
            additions['permissions'][role.uri] = permissions

    try:
        print_resources(cmd_ctx, user_roles, cmd_ctx.output_format, show_list,
                        additions, all=options['all'])
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)


def cmd_user_role_show(cmd_ctx, user_role_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user_role = find_user_role(cmd_ctx, console, user_role_name)

    try:
        user_role.pull_full_properties()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    properties = dict(user_role.properties)

    # Add artificial property 'parent-name'
    properties['parent-name'] = console.name

    # Improve representation of property 'permissions'
    obj_cache = ObjectByUriCache(cmd_ctx, client)
    permissions = []
    for item in user_role.properties['permissions']:
        permissions.append(permission_str(obj_cache, item))
    properties['permissions'] = permissions

    # Add artificial property 'associated-system-defined-user-role-name'
    urole_uri = user_role.get_property(
        'associated-system-defined-user-role-uri')
    if urole_uri:
        # User-defined roles have an associated system-defined role
        urole_props = client.session.get(urole_uri)
        urole_name = urole_props['name']
        properties['associated-system-defined-user-role-name'] = urole_name
    else:
        # System-defined roles do not have an associated system-defined role
        properties['associated-system-defined-user-role-name'] = None

    print_properties(cmd_ctx, properties, cmd_ctx.output_format)


def cmd_user_role_create(cmd_ctx, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console

    org_options = original_options(options)

    name_map = {
        'associated-system-defined-user-role': None,
    }

    properties = options_to_properties(org_options, name_map)

    if org_options['associated-system-defined-user-role'] == '':
        properties['associated-system-defined-user-role-uri'] = None
    else:
        try:
            role = console.user_roles.find_by_name(
                org_options['associated-system-defined-user-role'])
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
        properties['associated-system-defined-user-role-uri'] = role.uri

    try:
        new_user_role = console.user_roles.create(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("New user role '{u}' has been created.".
               format(u=new_user_role.properties['name']))


def cmd_user_role_update(cmd_ctx, user_role_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user_role = find_user_role(cmd_ctx, console, user_role_name)

    org_options = original_options(options)

    name_map = {
        'associated-system-defined-user-role': None,
    }

    properties = options_to_properties(org_options, name_map)

    if org_options['associated-system-defined-user-role'] == '':
        properties['associated-system-defined-user-role-uri'] = None
    else:
        try:
            role = console.user_roles.find_by_name(
                org_options['associated-system-defined-user-role'])
        except zhmcclient.Error as exc:
            raise click_exception(exc, cmd_ctx.error_format)
        properties['associated-system-defined-user-role-uri'] = role.uri

    if not properties:
        cmd_ctx.spinner.stop()
        click.echo("No properties specified for updating user role '{u}'.".
                   format(u=user_role_name))
        return

    try:
        user_role.update_properties(properties)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("User role '{u}' has been updated.".format(u=user_role_name))


def cmd_user_role_delete(cmd_ctx, user_role_name):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user_role = find_user_role(cmd_ctx, console, user_role_name)

    try:
        user_role.delete()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("User role '{u}' has been deleted.".format(u=user_role_name))


def cmd_user_role_add_permission(cmd_ctx, user_role_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user_role = find_user_role(cmd_ctx, console, user_role_name)

    kwargs = permission_options_to_kwargs(cmd_ctx, client, options)

    try:
        user_role.add_permission(**kwargs)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("The permission has been added to user role '{r}'.".
               format(r=user_role_name))


def cmd_user_role_remove_permission(cmd_ctx, user_role_name, options):
    # pylint: disable=missing-function-docstring

    client = zhmcclient.Client(cmd_ctx.session)
    console = client.consoles.console
    user_role = find_user_role(cmd_ctx, console, user_role_name)

    kwargs = permission_options_to_kwargs(cmd_ctx, client, options)

    try:
        user_role.remove_permission(**kwargs)
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    cmd_ctx.spinner.stop()
    click.echo("The permission has been removed from user role '{r}'.".
               format(r=user_role_name))



# Copyright 2016-2019 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for HMC info.
"""

from __future__ import absolute_import

import click

import zhmcclient
from .zhmccli import cli
from ._helper import print_properties, click_exception


@cli.command('info')
@click.pass_obj
def info(cmd_ctx):
    """
    Show information about the HMC.

    In addition to the command-specific options shown in this help text, the
    general options (see 'zhmc --help') can also be specified before the
    command.
    """
    cmd_ctx.execute_cmd(lambda: cmd_info(cmd_ctx))


def cmd_info(cmd_ctx):
    """
    Show information about the HMC.
    """
    client = zhmcclient.Client(cmd_ctx.session)
    try:
        api_version = client.query_api_version()
    except zhmcclient.Error as exc:
        raise click_exception(exc, cmd_ctx.error_format)

    print_properties(cmd_ctx, api_version, cmd_ctx.output_format)
